#!/bin/sh
#shellcheck disable=SC2039,SC1090

#
# Shell compatibility
#
# This should be the binary name of the shell we're running and is used to
# choose where to look for modules
#
SHELLFU_COMPAT=${SHELLFU_COMPAT:-$(ps -p$$ -ocmd= | cut -d' ' -f1 | rev | cut -d/ -f1 | rev)}

#
# Our installation directory
#
SHELLFU_DIR=${SHELLFU_DIR:-/usr/share/shellfu}

#
# Shellfu internals debug mode
#
SHELLFU_DEBUGINIT=${SHELLFU_DEBUGINIT:-false}

#
# Target to list imported functions in embedding mode
#
# In embedding mode, we need to log list of imported
# modules in an external file.
#
SHELLFU_EMBEDTMP=${SHELLFU_EMBEDTMP:-}

#
# True if running in embedded mode (imports do nothing)
#
SHELLFU_EMBEDDED=${SHELLFU_EMBEDDED:-false}

#
# Last git commit hash before build
#
SHELLFU_GIT_HASH=da22468e6c68bc898975d8ddfaabac5848dbc130

#
# Module folder prefix
#
# See shellfu() documentation for details about module loading.
#
SHELLFU_INCLUDE=${SHELLFU_INCLUDE:-$SHELLFU_DIR/include}

#
# Colon separated list of additional paths to search modules
#
# See shellfu() documentation for details about module loading.
#
SHELLFU_PATH=${SHELLFU_PATH:-}

#
# Shellfu version (SemVer 2.0)
#
SHELLFU_VERSION=0.10.62

#
# Colon-separated list of already imported modules
#
__SHELLFU_IMPORTED=${__SHELLFU_IMPORTED:-}

shellfu() {
    #
    # Shellfu init function
    #
    # Usage:
    #
    #     shellfu ACTION MODULE
    #
    # MODULE is selected and acted upon, depending on ACTION:
    #
    #  * `import` will source the module using shell built-in dot command.
    #
    #  * `try_import` will try to import the module in a separate environment
    #     and exit with zero on success (it will not modity the current
    #     environment).
    #
    #  *  `version` will look for directive `#shellfu module-version=VER`
    #     and print the VER part.  VER must not contain spaces.  Exit status
    #     is zero if a non-empty VER has been found.
    #
    # The module lookup algorithm triest to find file MODULE.sh in following
    # directories:
    #
    #  1. Any directories in colon-separated list SHELLFU_PATH,
    #
    #  2. directory formed by joining SHELLFU_INCLUDE and SHELLFU_COMPAT
    #     with dingle dash,
    #
    #  3. directory formed by joining 'sh' and SHELLFU_COMPAT with single
    #     dash.
    #
    # for example, following code:
    #
    #     #!/bin/bash
    #
    #     . $(sfpath)
    #
    #     SHELLFU_INCLUDE=/shellfu/include
    #     SHELLFU_PATH=/shellfu/addon1:/shellfu/addon2
    #
    #     shellfu import foo
    #
    # will import first existing file of these:
    #
    #     /shellfu/addon1/foo.sh
    #     /shellfu/addon2/foo.sh
    #     /shellfu/include-bash/foo.sh
    #     /shellfu/include-sh/foo.sh
    #
    # Note that SHELLFU_COMPAT should be auto-detected by shellfu and should
    # contain name of current shell binary being executed.  This means that
    # modules can exist in multiple versions for each supported shell, and
    # when loading, shellfu prefers the shell-specific implementation.
    #

    #shellcheck disable=SC3043
    case $SHELLFU_COMPAT in
        dash|bash|ksh|zsh)
            local __shellfu__cmd=
            local __shellfu__dbody=
            local __shellfu__dname=
            local __shellfu__dvalue=
            local __shellfu__es=
            local __shellfu__ex=
            local __shellfu__incdir=
            local __shellfu__init=
            local __shellfu__mname=
            local __shellfu__mpath=
            local __shellfu__msg=
            local __shellfu__s=
            local __shellfu__usage=
            ;;
    esac

    __shellfu__cmd=$1; shift
    __shellfu__msg=

    # debug on import
    case $SHELLFU_DEBUGINIT:$__shellfu__cmd in
        true:import)
            printf 'shellfu:debug:%s\n' >&2 \
                "\$*='$*'" \
                "SHELLFU_COMPAT='$SHELLFU_COMPAT'" \
                "SHELLFU_DIR='$SHELLFU_DIR'" \
                "SHELLFU_DEBUGINIT='$SHELLFU_DEBUGINIT'" \
                "SHELLFU_EMBEDDED='$SHELLFU_EMBEDDED'" \
                "SHELLFU_EMBEDTMP='$SHELLFU_EMBEDTMP'" \
                "SHELLFU_GIT_HASH='$SHELLFU_GIT_HASH'" \
                "__SHELLFU_IMPORTED='$__SHELLFU_IMPORTED'" \
                "SHELLFU_INCLUDE='$SHELLFU_INCLUDE'" \
                "SHELLFU_PATH='$SHELLFU_PATH'" \
                "SHELLFU_VERSION='$SHELLFU_VERSION'"
            ;;
        *)
            ;;
    esac

    case $__shellfu__cmd in
        __debug)
            $SHELLFU_DEBUGINIT || return 0
            for __shellfu__msg in "$@"; do
                echo "shellfu:debug: $__shellfu__msg" >&2
            done
            ;;
        __die)
            for __shellfu__msg in "$@"; do
                echo "shellfu:fatal: $__shellfu__msg" >&2
            done
            test -z "$PS1" && exit 3
            return 3
            ;;
        __warn)
            for __shellfu__msg in "$@"; do
                echo "shellfu: $__shellfu__msg" >&2
            done
            ;;
        __usage)
            echo "shellfu:usage: shellfu $1" >&2
            test -z "$PS1" && exit 3
            return 3
            ;;

        import)
            #
            # Import module named $1
            #
            __shellfu__mname=$1
            shellfu _is_imported "$__shellfu__mname" && return 0
            if shellfu _do_import "$__shellfu__mname"; then
                __SHELLFU_IMPORTED="$__SHELLFU_IMPORTED${__SHELLFU_IMPORTED:+:}$__shellfu__mname"
            else
                shellfu __die "cannot import module: $__shellfu__mname"
            fi
            ;;
        reload)
            #
            # Reload module named $1 (unsafe!)
            #
            __shellfu__mname=$1
            shellfu _is_imported "$__shellfu__mname" || {
                shellfu __warn "cannot reload; not imported yet: $__shellfu__mname"
                return 2
            }
            if ! shellfu _do_import "$__shellfu__mname"; then
                shellfu __die "cannot reload module: $__shellfu__mname"
            fi
            ;;
        try_import)
            #
            # Try if module $1 could be imported
            #
            __shellfu__mname=$1
            ( shellfu _do_import "$__shellfu__mname" )
            ;;
        version)
            #
            # Show version of module named $1
            #
            __shellfu__mname=$1
            __shellfu__mpath=
            __shellfu__mpath=$(shellfu _select_mfile "$__shellfu__mname") || {
                shellfu __warn "cannot find module: $__shellfu__mname"
                return 3
            }
            shellfu __debug "found module file: $__shellfu__mpath"
            shellfu _read_directive "module-version" "$__shellfu__mpath"
            ;;
        ls_imported)
            #
            # Print list of modules already imported
            #
            echo "$__SHELLFU_IMPORTED" | tr : \\n
            ;;

        _do_import)
            #
            # Really import module named $1
            #
            $SHELLFU_EMBEDDED && return 0
            __shellfu__mname=$1
            __shellfu__es=4
            __shellfu__mpath=
            shellfu __debug "importing module $__shellfu__mname"
            __shellfu__mpath=$(shellfu _select_mfile "$__shellfu__mname") || {
                shellfu __warn "cannot find module: $__shellfu__mname"
                return 3
            }
            shellfu __debug "found module file: $__shellfu__mpath"
            bash -n "$__shellfu__mpath" || return 3
            . "$__shellfu__mpath"
            __shellfu__es=$?
            test -n "$SHELLFU_EMBEDTMP" && echo "$__shellfu__mpath" >> "$SHELLFU_EMBEDTMP"
            __shellfu__init=__shellfu_${__shellfu__mname}__init
            if test "$(command -v "$__shellfu__init")" = "$__shellfu__init"; then
                shellfu __debug "launching init function: $__shellfu__init"
                $__shellfu__init
                __shellfu__es=$?
            fi
            return $__shellfu__es
            ;;
        _is_imported)
            #
            # True if module $1 is already imported
            #
            __shellfu__mname=$1
            echo "$__SHELLFU_IMPORTED" | tr : \\n | grep -qxe "$__shellfu__mname"
            ;;

        _list_mfiles)
            #
            # Find module files of module matching $1 (wildcard expression)
            #
            __shellfu__ex="${1:-*}"
            __shellfu__incdir=
            echo "$SHELLFU_PATH:$SHELLFU_INCLUDE-sh:$SHELLFU_INCLUDE-$SHELLFU_COMPAT" \
              | tr ":" '\n' \
              | awk '!seen[$0]++' \
              | while read -r __shellfu__incdir; do
                    shellfu __debug "checking: $__shellfu__incdir"
                    test -d "$__shellfu__incdir" || continue
                    shellfu __debug "looking into: $__shellfu__incdir"
                    find "$__shellfu__incdir" -name "$__shellfu__ex.sh"
                done
            ;;
        _select_mfile)
            #
            # Select file that contains module named $1
            #
            shellfu _list_mfiles "$1" | grep -m1 .
            ;;

        _read_directive)
            __shellfu__s=false
            test "$1" = "-s" && { __shellfu__s=true; shift; }
            __shellfu__dname=$1
            __shellfu__mpath=$2
            __shellfu__dbody=
            __shellfu__dvalue=
            __shellfu__usage="_read_directive DIRECTIVE PATH/TO/MODULE.sh"
            test -n "$__shellfu__dname" || shellfu __usage "$__shellfu__usage"
            test -n "$__shellfu__mpath" || shellfu __usage "$__shellfu__usage"
            test -f "$__shellfu__mpath" || shellfu __die "no such file: $__shellfu__mpath"
            shellfu __debug "reading directive: '$__shellfu__dname' from '$__shellfu__mpath'"
            __shellfu__dbody=$(
                { head -3 "$__shellfu__mpath"; tail -3 "$__shellfu__mpath"; } \
                  | grep '^#shellfu ' \
                  | tr ' ' '\n' \
                  | grep -m1 "^$__shellfu__dname=[^ ]*"
            )
            test -n "$__shellfu__dbody" || {
                $__shellfu__s || shellfu __warn "directive not found: '$__shellfu__dname' in '$__shellfu__mpath'"
                return 2
            }
            __shellfu__dvalue=${__shellfu__dbody#"$__shellfu__dname="}
            printf %s "$__shellfu__dvalue"
            test -n "$__shellfu__dvalue"
            ;;

        *)
            shellfu __die "unknown shellfu command: $__shellfu__cmd"
            ;;
    esac
}
